# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import numpy as np
import torch

from ptbench.engine.saliencymap_evaluator import (
    _compute_avg_saliency_focus,
    _compute_max_iou_and_ioda,
    _compute_proportional_energy,
    _compute_simultaneous_iou_and_ioda,
    calculate_localization_metrics,
)


def test_compute_max_iou_and_ioda():
    detected_box = (10, 10, 100, 100)
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
    ]
    gt_box_dict2 = [
        torch.tensor(1),
        torch.tensor(20),
        torch.tensor(20),
        torch.tensor(60),
        torch.tensor(60),
    ]
    gt_boxes = [gt_box_dict, gt_box_dict2]

    iou, ioda = _compute_max_iou_and_ioda(detected_box, gt_boxes)

    expected_iou = 0.36
    expected_ioda = 0.36

    assert iou == expected_iou
    assert ioda == expected_ioda


def test_compute_max_iou_and_ioda_zero_detected_area():
    detected_box = (10, 10, 0, 0)
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
    ]
    gt_boxes = [gt_box_dict]

    iou, ioda = _compute_max_iou_and_ioda(detected_box, gt_boxes)

    # Should be zero as the detected box has no area
    assert iou == 0
    assert ioda == 0


def test_compute_max_iou_and_ioda_zero_gt_area():
    detected_box = (10, 10, 100, 100)
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(0),
        torch.tensor(0),
    ]
    gt_boxes = [gt_box_dict]

    iou, ioda = _compute_max_iou_and_ioda(detected_box, gt_boxes)

    # Should be zero as there is no ground truth box
    assert iou == 0
    assert ioda == 0


def test_compute_max_iou_and_ioda_zero_intersection():
    detected_box = (10, 10, 100, 100)
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(0),
        torch.tensor(0),
        torch.tensor(5),
        torch.tensor(5),
    ]
    gt_boxes = [gt_box_dict]

    iou, ioda = _compute_max_iou_and_ioda(detected_box, gt_boxes)

    assert iou == 0
    assert ioda == 0


def test_compute_simultaneous_iou_and_ioda():
    detected_box = (10, 10, 100, 100)
    gt_box_dict1 = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
    ]
    gt_box_dict2 = [
        torch.tensor(1),
        torch.tensor(70),
        torch.tensor(70),
        torch.tensor(30),
        torch.tensor(30),
    ]
    gt_boxes = [gt_box_dict1, gt_box_dict2]

    iou, ioda = _compute_simultaneous_iou_and_ioda(detected_box, gt_boxes)

    assert iou == 0.34
    assert ioda == 0.34


def test_compute_avg_saliency_focus():
    grayscale_cams = np.ones((200, 200))
    grayscale_cams2 = np.full((512, 512), 0.5)
    grayscale_cams3 = np.zeros((256, 256))
    grayscale_cams3[50:75, 50:100] = 1
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
    ]
    gt_boxes = [gt_box_dict]

    avg_saliency_focus = _compute_avg_saliency_focus(gt_boxes, grayscale_cams)
    avg_saliency_focus2 = _compute_avg_saliency_focus(gt_boxes, grayscale_cams2)
    avg_saliency_focus3 = _compute_avg_saliency_focus(gt_boxes, grayscale_cams3)

    assert avg_saliency_focus == 1
    assert avg_saliency_focus2 == 0.5
    assert avg_saliency_focus3 == 0.5


def test_compute_avg_saliency_focus_no_activations():
    grayscale_cams = np.zeros((200, 200))
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
    ]
    gt_boxes = [gt_box_dict]

    avg_saliency_focus = _compute_avg_saliency_focus(gt_boxes, grayscale_cams)

    assert avg_saliency_focus == 0


def test_compute_avg_saliency_focus_zero_gt_area():
    grayscale_cams = np.ones((200, 200))
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(0),
        torch.tensor(0),
    ]
    gt_boxes = [gt_box_dict]

    avg_saliency_focus = _compute_avg_saliency_focus(gt_boxes, grayscale_cams)

    assert avg_saliency_focus == 0


def test_compute_proportional_energy():
    grayscale_cams = np.ones((200, 200))
    grayscale_cams2 = np.full((512, 512), 0.5)
    grayscale_cams3 = np.zeros((512, 512))
    grayscale_cams3[100:200, 100:200] = 1
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(100),
        torch.tensor(100),
    ]
    gt_boxes = [gt_box_dict]

    proportional_energy = _compute_proportional_energy(gt_boxes, grayscale_cams)
    proportional_energy2 = _compute_proportional_energy(
        gt_boxes, grayscale_cams2
    )
    proportional_energy3 = _compute_proportional_energy(
        gt_boxes, grayscale_cams3
    )

    assert proportional_energy == 0.25
    assert proportional_energy2 == 0.03814697265625
    assert proportional_energy3 == 0.25


def test_compute_proportional_energy_no_activations():
    grayscale_cams = np.zeros((200, 200))
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
    ]
    gt_boxes = [gt_box_dict]

    proportional_energy = _compute_proportional_energy(gt_boxes, grayscale_cams)

    assert proportional_energy == 0


def test_compute_proportional_energy_no_gt_box():
    grayscale_cams = np.ones((200, 200))
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(0),
        torch.tensor(0),
        torch.tensor(0),
        torch.tensor(0),
    ]
    gt_boxes = [gt_box_dict]

    proportional_energy = _compute_proportional_energy(gt_boxes, grayscale_cams)

    assert proportional_energy == 0


def test_calculate_localization_metrics():
    grayscale_cams = np.zeros((200, 200))
    detected_box = (10, 10, 100, 100)
    gt_box_dict = [
        torch.tensor(1),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
        torch.tensor(50),
    ]
    ground_truth_box = [gt_box_dict]

    (
        iou,
        ioda,
        proportional_energy,
        avg_saliency_focus,
    ) = calculate_localization_metrics(
        grayscale_cams, detected_box, ground_truth_box
    )

    assert iou == 0.25
    assert ioda == 0.25
    assert proportional_energy == 0
    assert avg_saliency_focus == 0
